@@ -34,7 +34,7 @@
 #include "xalloc.h"		/* Gnulib */
 #include "intutil.h"
 #include "gmediaserver.h"
-#include "web/MediaServer.h"
+#include "schemas/MediaServer.h"
 
 char *friendly_name = NULL;
 static UpnpDevice_Handle device;
@@ -410,6 +410,9 @@ init_upnp(const char *listenip, uint16_t listenport)
     if (res != UPNP_E_SUCCESS)
         die(_("Cannot set virtual directory callbacks - %s\n"), upnp_errmsg(res));
     res = UpnpAddVirtualDir("/audio");
+    if (res != UPNP_E_SUCCESS)
+        die(_("Cannot add virtual directory for web server - %s\n"), upnp_errmsg(res));
+    res = UpnpAddVirtualDir("/upnp");
     if (res != UPNP_E_SUCCESS)
         die(_("Cannot add virtual directory for web server - %s\n"), upnp_errmsg(res));
 
