@@ -29,11 +29,26 @@
 #include "xalloc.h"
 #include "intutil.h"
 #include "gmediaserver.h"
+#include "schemas/ConnectionManager.h"
+#include "schemas/ContentDirectory.h"
 
 typedef struct {
-    int fd;
     char *fullpath;
-    Entry *entry;
+    size_t pos;
+    enum {
+	FILE_LOCAL,
+	FILE_MEMORY,
+    } type;
+    union {
+	struct {
+	    int fd;
+	    Entry *entry;
+	} local;
+	struct {
+	    const char *contents;
+	    size_t len;
+	} memory;
+    } detail;
 } WebServerFile;
 
 static Entry *
@@ -64,6 +79,23 @@ webserver_get_info(const char *filename, struct File_Info *info)
 
     say(4, _("Request file information for %s\n"), filename);
 
+    if (strcmp(filename, "/upnp/ContentDirectory.xml") == 0) {
+	info->file_length = CONTENTDIRECTORY_DESC_LEN;
+	info->last_modified = 0;
+	info->is_directory = 0;
+	info->is_readable = 1;
+	info->content_type = ixmlCloneDOMString("text/xml");
+	return 0;
+    }
+    if (strcmp(filename, "/upnp/ConnectionManager.xml") == 0) {
+	info->file_length = CONNECTIONMANAGER_DESC_LEN;
+	info->last_modified = 0;
+	info->is_directory = 0;
+	info->is_readable = 1;
+	info->content_type = ixmlCloneDOMString("text/xml");
+	return 0;
+    }
+
     entry = get_file_entry(filename);
     if (entry == NULL)
 	return -1;
@@ -114,6 +146,25 @@ webserver_open(const char *filename, enum UpnpOpenFileMode mode)
 	return NULL;
     }
 
+    if (strcmp(filename, "/upnp/ContentDirectory.xml") == 0) {
+	file = xmalloc(sizeof(WebServerFile));
+	file->fullpath = xstrdup("ContentDirectory.xml in memory");
+	file->pos = 0;
+	file->type = FILE_MEMORY;
+	file->detail.memory.contents = CONTENTDIRECTORY_DESC;
+	file->detail.memory.len = CONTENTDIRECTORY_DESC_LEN;
+	return file;
+    }
+    if (strcmp(filename, "/upnp/ConnectionManager.xml") == 0) {
+	file = xmalloc(sizeof(WebServerFile));
+	file->fullpath = xstrdup("ConnectionManager.xml in memory");
+	file->pos = 0;
+	file->type = FILE_MEMORY;
+	file->detail.memory.contents = CONNECTIONMANAGER_DESC;
+	file->detail.memory.len = CONNECTIONMANAGER_DESC_LEN;
+	return file;
+    }
+
     entry = get_file_entry(filename);
     if (entry == NULL)
 	return NULL;
@@ -128,9 +179,11 @@ webserver_open(const char *filename, enum UpnpOpenFileMode mode)
     }
 
     file = xmalloc(sizeof(WebServerFile));
-    file->entry = entry;
-    file->fd = fd;
     file->fullpath = fullpath;
+    file->pos = 0;
+    file->type = FILE_LOCAL;
+    file->detail.local.entry = entry;
+    file->detail.local.fd = fd;
 
     return (UpnpWebFileHandle) file;
 }
@@ -139,13 +192,24 @@ static int
 webserver_read(UpnpWebFileHandle fh, char *buf, size_t buflen)
 {
     WebServerFile *file = (WebServerFile *) fh;
-    ssize_t len;
+    ssize_t len = -1;
 
-    say(4, _("Attempting to read %d bytes from %s\n"), buflen, file->fullpath);
+    say(4, _("Attempting to read %d bytes at %d from %s\n"), buflen, file->pos, file->fullpath);
+
+    switch (file->type) {
+    case FILE_LOCAL:
+	len = read(file->detail.local.fd, buf, buflen);
+	break;
+    case FILE_MEMORY:
+	len = min(buflen, file->detail.memory.len - file->pos);
+	memcpy(buf, file->detail.memory.contents + file->pos, len);
+	break;
+    }
 
-    len = read(file->fd, buf, buflen);
     if (len < 0)
 	warn(_("%s: cannot read: %s\n"), file->fullpath, errstr);
+    else
+	file->pos += len;
 
     return len;
 }
@@ -164,14 +228,56 @@ static int
 webserver_seek(UpnpWebFileHandle fh, long offset, int origin)
 {
     WebServerFile *file = (WebServerFile *) fh;
+    long newpos = -1;
 
-    say(4, _("Attempting to seek to %ld (origin %d) in %s\n"), offset, origin, file->fullpath);
-
-    if (lseek(file->fd, offset, origin) == -1) {
-	warn(_("%s: cannot seek: %s\n"), file->fullpath, errstr);
-	return -1;
+    switch (origin) {
+    case SEEK_SET:
+	say(4, _("Attempting to seek to %ld (was at %d) in %s\n"), offset, file->pos, file->fullpath);
+	newpos = offset;
+	break;
+    case SEEK_CUR:
+	say(4, _("Attempting to seek by %ld from %d in %s\n"), offset, file->pos, file->fullpath);
+	newpos = file->pos + offset;
+	break;
+    case SEEK_END:
+	say(4, _("Attempting to seek by %ld from end (was at %d) in %s\n"), offset, file->pos, file->fullpath);
+	if (file->type == FILE_LOCAL) {
+	    struct stat sb;
+	    if (stat(file->fullpath, &sb) < 0) {
+		warn(_("%s: cannot stat: %s\n"), file->fullpath, errstr);
+		return -1;
+	    }
+	    newpos = sb.st_size + offset;
+	} else if (file->type == FILE_MEMORY) {
+	    newpos = file->detail.memory.len + offset;
+	}
+	break;
     }
 
+    switch (file->type) {
+    case FILE_LOCAL:
+	/* Just make sure we cannot seek before start of file. */
+	if (newpos < 0) {
+	    warn(_("%s: cannot seek: %s\n"), file->fullpath, strerror(EINVAL));
+	    return -1;
+	}
+	/* Don't seek with origin as specified above, as file may have
+	 * changed in size since our last stat.
+	 */
+	if (lseek(file->detail.local.fd, newpos, SEEK_SET) == -1) {
+	    warn(_("%s: cannot seek: %s\n"), file->fullpath, errstr);
+	    return -1;
+	}
+	break;
+    case FILE_MEMORY:
+	if (newpos < 0 || newpos > file->detail.memory.len) {
+	    warn(_("%s: cannot seek: %s\n"), file->fullpath, strerror(EINVAL));
+	    return -1;
+	}
+	break;
+    }
+
+    file->pos = newpos;
     return 0;
 }
 
@@ -180,15 +286,21 @@ webserver_close(UpnpWebFileHandle fh)
 {
     WebServerFile *file = (WebServerFile *) fh;
 
-    close(file->fd); /* Ignore errors since only reading */
+    switch (file->type) {
+    case FILE_LOCAL:
+	close(file->detail.local.fd); /* Ignore errors since only reading */
+	break;
+    case FILE_MEMORY:
+	/* no operation */
+	break;
+    }
     free(file->fullpath);
     free(file);
 
     return 0;
 }
 
-struct UpnpVirtualDirCallbacks virtual_dir_callbacks =
-{
+struct UpnpVirtualDirCallbacks virtual_dir_callbacks = {
     webserver_get_info,
     webserver_open,
     webserver_read,
